use super::*;

/// Preferred way to parse [Point]s off of a [Serializer].
pub struct Addresses {
    current: usize,
}

impl Addresses {
    /// Read the next [Address].
    pub fn next<'a>(
        &mut self,
        deserializer: &'a mut dyn Deserializer,
    ) -> Result<Address, &'a [u8]> {
        let point = deserializer.read_n_const::<HASH_SIZE>()?;
        let address = Address {
            point,
            index: self.current,
        };
        self.current += 1;
        Ok(address)
    }

    /// Read the next [Point].
    pub fn next_point<'a, 'b, Ctx: 'a + Context, A: Mentionable<'a, Ctx>>(
        &mut self,
        deserializer: &'b mut dyn Deserializer,
        resolver: Rc<dyn Resolver<'a, Ctx>>,
        factory: A::Fctr,
    ) -> Result<Point<'a, Ctx, A>, &'b [u8]> {
        Ok(Point::from_address(
            self.next(deserializer)?,
            factory,
            resolver,
        ))
    }

    /// Start reading the [Address]es.
    fn start() -> Self {
        Addresses { current: 0 }
    }
}

fn _parse_slice<'a, Ctx: 'a + Context, A: Mentionable<'a, Ctx>>(
    factory: &A::Fctr,
    slice: &[u8],
    resolver: Rc<dyn Resolver<'a, Ctx>>,
) -> ParseResult<'a, Ctx, A::Fctr> {
    let mut deserializer = SliceDeserializer::from(slice);
    let mentionable = factory.deserialize(&mut deserializer, resolver, &mut Addresses::start())?;
    let tail = deserializer.read_all();
    if tail.is_empty() {
        Ok(mentionable)
    } else {
        Err(factory.unexpected_tail(tail))
    }
}

impl<'a, Ctx: 'a + Context, F: Factory<'a, Ctx>> ExtFactory<'a, Ctx> for F {
    fn parse_slice(
        &self,
        slice: &[u8],
        resolver: Rc<dyn Resolver<'a, Ctx>>,
    ) -> ParseResult<'a, Ctx, Self> {
        _parse_slice::<Ctx, F::Mtbl>(self, slice, resolver)
    }
}
